import asyncio


from mikro_next.scalars import ArrayLike, ImageFileLike, MeshLike, ParquetLike, FileLike
from rath.links.parsing import ParsingLink
from rath.operation import Operation, opify
from typing import Any, Tuple, Type, Union
from mikro_next.io.upload import (
    aupload_bigfile,
    aupload_xarray,
    aupload_parquet,
    astore_media_file,
    astore_mesh_file,
)
from pydantic import Field
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from mikro_next.datalayer import DataLayer
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mikro_next.api.schema import Credentials, PresignedPostCredentials
    from mikro_next.datalayer import DataLayer
    from mikro_next.io.upload import FileLike, MeshLike, ParquetLike, ArrayLike, ImageFileLike


async def apply_recursive(func, obj, typeguard: Union[Type[Any], Tuple[Type[Any], ...]]) -> Any:  # type: ignore
    """
    Recursively applies an asynchronous function to elements in a nested structure.

    Args:
        func (callable): The asynchronous function to apply.
        obj (any): The nested structure (dict, list, tuple, etc.) to process.
        typeguard (type): The type of elements to apply the function to.

    Returns:
        any: The nested structure with the function applied to elements of the specified type.
    """
    if isinstance(obj, dict):  # If obj is a dictionary, recursively apply to each key-value pair
        return {k: await apply_recursive(func, v, typeguard) for k, v in obj.items()}  # type: ignore
    elif isinstance(obj, list):  # If obj is a list, recursively apply to each element
        return await asyncio.gather(*[apply_recursive(func, elem, typeguard) for elem in obj])  # type: ignore
    elif isinstance(
        obj, tuple
    ):  # If obj is a tuple, recursively apply to each element and convert back to tuple
        return tuple(
            await asyncio.gather(*[apply_recursive(func, elem, typeguard) for elem in obj])  # type: ignore
        )
    elif isinstance(obj, typeguard):
        return await func(obj)  # type: ignore
    else:  # If obj is not a dict, list, tuple, or matching the typeguard, return it as is
        return obj  # type: ignore


class UploadLink(ParsingLink):
    """Data Layer Upload Link

    This link is used to upload  supported types to a DataLayer.
    It parses queries, mutatoin and subscription arguments and
    uploads the items to the DataLayer, and substitures the
    DataFrame with the S3 path.

    Args:
        ParsingLink (_type_): _description_


    """

    datalayer: DataLayer

    executor: ThreadPoolExecutor = Field(
        default_factory=lambda: ThreadPoolExecutor(max_workers=4), exclude=True
    )
    _executor_session: Any = None

    async def __aenter__(self) -> "UploadLink":
        """Enter the context manager for the UploadLink"""
        self._executor_session = self.executor.__enter__()
        return self

    async def aget_image_credentials(self, key: str, datalayer: str) -> "Credentials":
        """Get image upload credentials"""
        from mikro_next.api.schema import RequestUploadMutation, RequestUploadInput

        operation = opify(
            RequestUploadMutation.Meta.document,
            variables={"input": RequestUploadInput(key=key, datalayer=datalayer).model_dump()},
        )

        if not self.next:
            raise ValueError("No next link found. Please set the next link.")

        async for result in self.next.aexecute(operation):
            return RequestUploadMutation(**result.data).request_upload

        raise ValueError("No result found for image upload credentials")

    async def aget_table_credentials(self, key: str, datalayer: str) -> "Credentials":
        """Get table upload credentials"""
        from mikro_next.api.schema import (
            RequestTableUploadMutation,
            RequestTableUploadInput,
        )

        if not self.next:
            raise ValueError("No next link found. Please set the next link.")

        operation = opify(
            RequestTableUploadMutation.Meta.document,
            variables={
                "input": RequestTableUploadInput(key=key, datalayer=datalayer).model_dump(
                    by_alias=True, exclude_unset=True
                )
            },
        )

        async for result in self.next.aexecute(operation):
            return RequestTableUploadMutation(**result.data).request_table_upload

        raise ValueError("No result found for table upload credentials")

    async def aget_bigfile_credentials(self, file: FileLike, datalayer: str) -> Any:
        from mikro_next.api.schema import (
            RequestFileUploadMutation,
            RequestFileUploadInput,
        )

        if not self.next:
            raise ValueError("No next link found. Please set the next link.")

        operation = opify(
            RequestFileUploadMutation.Meta.document,
            variables={
                "input": RequestFileUploadInput(
                    fileName=file.file_name, datalayer=datalayer
                ).model_dump(by_alias=True, exclude_unset=True)
            },
        )

        async for result in self.next.aexecute(operation):
            return RequestFileUploadMutation(**result.data).request_file_upload

    async def aget_mesh_credentials(self, key: str, datalayer: str) -> "PresignedPostCredentials":
        from mikro_next.api.schema import (
            RequestMeshUploadMutation,
            RequestMeshUploadInput,
        )

        if not self.next:
            raise ValueError("No next link found. Please set the next link.")

        operation = opify(
            RequestMeshUploadMutation.Meta.document,
            variables={
                "input": RequestMeshUploadInput(key=key, datalayer=datalayer).model_dump(
                    by_alias=True, exclude_unset=True
                )
            },
        )

        async for result in self.next.aexecute(operation):
            return RequestMeshUploadMutation(**result.data).request_mesh_upload

        raise ValueError("No result found for mesh upload credentials")

    async def arequest_media_credentials(
        self, file_name: str, datalayer: str
    ) -> "PresignedPostCredentials":
        from mikro_next.api.schema import (
            RequestMediaUploadMutation,
            RequestMediaUploadInput,
        )

        if not self.next:
            raise ValueError("No next link found. Please set the next link.")

        operation = opify(
            RequestMediaUploadMutation.Meta.document,
            variables={
                "input": RequestMediaUploadInput(
                    fileName=file_name, datalayer=datalayer
                ).model_dump(by_alias=True, exclude_unset=True)
            },
        )

        async for result in self.next.aexecute(operation):
            return RequestMediaUploadMutation(**result.data).request_media_upload

        raise ValueError("No result found for media upload credentials")

    async def aupload_parquet(self, datalayer: "DataLayer", parquet_input: ParquetLike) -> str:
        """Upload a Parquet file to the DataLayer asynchronously."""
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_table_credentials(parquet_input.key, endpoint_url)
        return await aupload_parquet(
            parquet_input,
            credentials,
            datalayer,
            self._executor_session,
        )

    async def aupload_xarray(self, datalayer: "DataLayer", xarray: ArrayLike) -> str:
        """Upload an xarray to the DataLayer asynchronously."""
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_image_credentials(xarray.key, endpoint_url)
        return await aupload_xarray(
            xarray,
            credentials,
            datalayer,
        )

    async def aupload_bigfile(self, datalayer: "DataLayer", file: FileLike) -> str:
        """Upload a big file to the DataLayer asynchronously."""
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_bigfile_credentials(file, endpoint_url)
        return await aupload_bigfile(
            file,
            credentials,
            datalayer,
        )

    async def aupload_mediafile(self, datalayer: "DataLayer", file: ImageFileLike) -> str:
        """Upload a media file to the DataLayer asynchronously."""
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.arequest_media_credentials(file.file_name, endpoint_url)
        return await astore_media_file(
            file,
            credentials,
            datalayer,
        )

    async def astore_mesh_file(self, datalayer: "DataLayer", mesh: MeshLike) -> str:
        """Store a mesh file in the DataLayer asynchronously."""
        assert datalayer is not None, "Datalayer must be set"
        endpoint_url = await datalayer.get_endpoint_url()

        credentials = await self.aget_mesh_credentials(mesh.key, endpoint_url)
        return await astore_mesh_file(
            mesh,
            credentials,
            datalayer,
        )

    async def aparse(self, operation: Operation) -> Operation:
        """Parse the operation (Async)

        Extracts the DataFrame from the operation and uploads it to the DataLayer.

        Args:
            operation (Operation): The operation to parse

        Returns:
            Operation: _description_
        """

        datalayer = operation.context.kwargs.get("datalayer", self.datalayer)

        operation.variables = await apply_recursive(
            partial(self.aupload_xarray, datalayer),
            operation.variables,
            ArrayLike,
        )
        operation.variables = await apply_recursive(
            partial(self.aupload_parquet, datalayer), operation.variables, ParquetLike
        )
        operation.variables = await apply_recursive(
            partial(self.aupload_bigfile, datalayer),
            operation.variables,
            (FileLike),
        )
        operation.variables = await apply_recursive(
            partial(self.aupload_mediafile, datalayer),
            operation.variables,
            (ImageFileLike),
        )
        operation.variables = await apply_recursive(
            partial(self.astore_mesh_file, datalayer), operation.variables, MeshLike
        )

        return operation

    async def adisconnect(self) -> None:
        """Disconnect the UploadLink"""
        self.executor.__exit__(None, None, None)



"""
Custom scalars for mikro_next


"""

import io
import os
from typing import Any, IO, List, Optional, TypeAlias, cast
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema
import xarray as xr
import pandas as pd
import numpy as np
import uuid
from numpy.typing import NDArray

from .utils import rechunk
from collections.abc import Iterable
import mimetypes
from pathlib import Path
from pydantic_core import core_schema


OneDArray = NDArray[np.generic]
TwoDArray = NDArray[np.generic]


TwoDVectorCoercible: TypeAlias = List[float] | OneDArray | List[int]
""" A type alias for 2D vector-like structures that can be coerced into a TwoDVector."""

ThreeDVectorCoercible: TypeAlias = List[float] | OneDArray | List[int]
""" A type alias for 3D vector-like structures that can be coerced into a ThreeDVector."""

FourDVectorCoercible: TypeAlias = List[float] | OneDArray | List[int]
""" A type alias for 4D vector-like structures that can be coerced into a FourDVector."""

FiveDVectorCoercible: TypeAlias = List[float] | OneDArray | List[List[float]] | List[List[int]]
""" A type alias for 5D vector-like structures that can be coerced into a FiveDVector."""

ArrayCoercible: TypeAlias = xr.DataArray | OneDArray | List[float] | List[List[float]]
""" A type alias for array-like structures that can be coerced into an xarray DataArray."""

ImageFileCoercible: TypeAlias = str | bytes | Path | io.BufferedReader
""" A type alias for image file-like structures that can be coerced into an xarray DataArray."""

ParquetCoercible: TypeAlias = pd.DataFrame
""" A type alias for parquet-like structures that can be coerced into an xarray DataArray."""

MeshCoercible: TypeAlias = str | bytes | Path | io.BufferedReader
""" A type alias for mesh-like structures that can be coerced into an xarray DataArray."""

FileCoercible: TypeAlias = str | bytes | Path | io.BufferedReader
""" A type alias for file-like structures that can be coerced into an xarray DataArray."""

FourByFourMatrixCoercible: TypeAlias = list[list[float]] | TwoDArray | list[list[int]]
""" A type alias for 4x4 matrix-like structures that can be coerced into an xarray DataArray."""

MillisecondsCoercible: TypeAlias = int | float
""" A type alias for millisecond-like structures that can be coerced into an xarray DataArray."""

MicrometersCoercible: TypeAlias = int | float
""" A type alias for micrometer-like structures that can be coerced into an xarray DataArray."""

RGBAColorCoercible: TypeAlias = List[float] | List[int] | OneDArray
""" A type alias for RGBA color-like structures that can be coerced into an RGBA Value"""


def is_dask_array(v: Any) -> bool:  # noqa: ANN401
    """Check if the input is a dask array."""
    try:
        import dask.array.core as da

        return isinstance(v, da.Array)
    except ImportError:
        return False
    except Exception as e:
        raise ValueError(f"Error checking for dask array: {e}")


class RGBAColor(list[float]):
    """A custom scalar to represent an affine matrix."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(float))

    @classmethod
    def validate(cls, v: RGBAColorCoercible) -> "RGBAColor":
        """Validate the input array and convert it to a xr.DataArray."""
        if isinstance(v, np.ndarray):
            if v.ndim == 1:
                v = v.tolist()
            else:
                raise ValueError("The input array must be a 1D array")

        if not isinstance(v, list):
            raise ValueError("The input must be a list or a 1-D numpy array.")

        v = [float(i) for i in v]  # Convert all elements to float

        if len(v) == 3:
            v.append(1.0)  # Add alpha channel if not present

        if len(v) != 4:
            raise ValueError(
                f"The input must be a list of 3 or 4 elements (R, G, B, [A]). You provided a list of {len(v)} elements"
            )

        return cls(v)


class XArrayConversionException(Exception):
    """An exception that is raised when a conversion to xarray fails."""

    pass


MetricValue = Any
FeatureValue = Any


class Micrometers(float):
    """A custom scalar to represent a micrometer."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(float))

    @classmethod
    def validate(cls, v: MicrometersCoercible) -> "Micrometers":
        """Validate the input array and convert it to a xr.DataArray."""
        return cls(v)


class Milliseconds(float):
    """A custom scalar to represent a millisecond."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(float))

    @classmethod
    def validate(cls, v: MillisecondsCoercible) -> "Milliseconds":
        """Validate the input array and convert it to a xr.DataArray."""
        return cls(v)


class TwoDVector(list[float]):
    """A custom scalar to represent a vector."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(list))

    @classmethod
    def validate(cls, v: TwoDVectorCoercible) -> "TwoDVector":
        """Validate the input array and convert it to a xr.DataArray."""
        if isinstance(v, np.ndarray):
            assert v.ndim == 1
            v = v.tolist()  # Convert numpy array to list #type: ignore

        assert isinstance(v, list)
        assert len(v) == 3
        return cls(v)

    def as_vector(self) -> OneDArray:
        """Convert the TwoDVector to a numpy array."""
        return np.array(self)


class ThreeDVector(list[float]):
    """A custom scalar to represent a vector."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(list))

    @classmethod
    def validate(cls, v: ThreeDVectorCoercible) -> "ThreeDVector":
        """Validate the input array and convert it to a xr.DataArray."""
        if isinstance(v, np.ndarray):
            assert v.ndim == 1
            v = v.tolist()

        assert isinstance(v, list)
        assert len(v) == 3
        return cls(v)

    def as_vector(self) -> OneDArray:
        """Convert the ThreeDVector to a numpy array."""
        return np.array(self)


class FourDVector(list[float]):
    """A custom scalar to represent a vector."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(list))

    @classmethod
    def validate(cls, v: FourDVectorCoercible) -> "FourDVector":
        """Validate the input array and convert it to a xr.DataArray."""
        if isinstance(v, np.ndarray):
            assert v.ndim == 1
            v = v.tolist()

        assert isinstance(v, list)
        assert len(v) == 4
        return cls(v)

    def as_vector(self) -> OneDArray:
        """Convert the FourDVector to a numpy array."""
        return np.array(self).reshape(-1)


class FiveDVector(list[float]):
    """A custom scalar to represent a vector."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(list))

    @classmethod
    def validate(cls, v: FiveDVectorCoercible) -> "FiveDVector":
        """Validate the input array and convert it to a xr.DataArray."""

        if isinstance(v, np.ndarray):
            if not v.ndim == 1:
                raise ValueError("The input array must be a 1D array")
            v = v.tolist()

        if not isinstance(v, Iterable):  # type: ignore
            raise ValueError("The input must be a list or a 1-D numpy array.")

        if not isinstance(v, list):
            v = list(v)

        for i in v:
            if not isinstance(i, (int, float)):
                raise ValueError(
                    f"The input must be a list of integers or floats. You provided a list of {type(i)}"
                )

        if len(v) < 2 or len(v) > 5:
            raise ValueError(
                f"The input must be a list or at least 2 elements (x, y) but not more than 5e lements (c, t, z, x, y). Every additional element is a z value (c, t, z, x, y). You provided a list o {len(v)} elements"
            )

        return v  # type: ignore

    @classmethod
    def list_from_numpyarray(
        cls,
        x: OneDArray,
        t: Optional[int] = None,
        c: Optional[int] = None,
        z: Optional[int] = None,
    ) -> List["FiveDVector"]:
        """Creates a list of FiveDVectors from a numpy array

        Args:
            vector_list (List[List[float]]): A list of lists of floats

        Returns:
            List[Vectorizable]: A list of InputVector
        """
        assert x.ndim == 2, "Needs to be a List array of vectors"
        if x.shape[1] == 4:
            return [FiveDVector([c] + i) for i in x.tolist()]
        elif x.shape[1] == 3:
            return [FiveDVector([c, t] + i) for i in x.tolist()]
        elif x.shape[1] == 2:
            return [FiveDVector([c, t, z] + i) for i in x.tolist()]
        else:
            raise NotImplementedError(
                f"Incompatible shape {x.shape} of {x}. List dimension needs to either be of size 2 or 3"
            )

    def as_vector(self) -> OneDArray:
        """Convert the FiveDVector to a numpy array."""
        return np.array(self)


class FourByFourMatrix(list[list[float]]):
    """A custom scalar to represent a four by four matrix (e.g 3D affine matrix.)"""

    def __get__(self, instance, owner) -> "FourByFourMatrix": ...  # type: ignore # noqa: ANN001, D105

    def __set__(self, instance, value: FourByFourMatrixCoercible) -> None: ...  # type: ignore # noqa: ANN001, D105

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_before_validator_function(cls.validate, handler(list))

    @classmethod
    def validate(cls, v: FourByFourMatrixCoercible) -> "FourByFourMatrix":
        """Validate the input array and convert it to a xr.DataArray."""
        if isinstance(v, np.ndarray):
            if not v.ndim == 2:
                raise ValueError("The input array must be a 2D array")
            if not v.shape[0] == v.shape[1]:
                raise ValueError("The input array must be a square matrix")
            if not v.shape == (4, 4):
                raise ValueError("The input array must be a 4x4 matrix")
            clean = [[float(v[i, j]) for j in range(4)] for i in range(4)]
        else:
            clean = v

        if not isinstance(clean, list):  # type: ignore
            raise ValueError(
                f"Expected a list or numpy array, got {type(clean)}. Please provide a 4x4 matrix."
            )

        if len(clean) != 4 or any(len(row) != 4 for row in clean):
            raise ValueError(
                f"Expected a 4x4 matrix, got {len(clean)} rows and {[len(row) for row in clean]} columns."
            )

        for row in clean:
            if not all(isinstance(x, (int, float)) for x in row):  # type: ignore
                raise ValueError("All elements of the 4x4 matrix must be integers or floats.")

        print(f"Validating FourByFourMatrix: {clean}")
        return cls(clean)  # type: ignore

    def as_matrix(self) -> TwoDArray:
        """Convert the FourByFourMatrix to a numpy array."""
        return np.array(self).reshape(4, 4)

    @classmethod
    def from_np(cls, v: TwoDArray) -> "FourByFourMatrix":
        """Validate the input array and convert it to a xr.DataArray."""
        return cls.validate(v)


class ArrayLike:
    """A custom scalar for wrapping of every supported array like structure on
    the mikro platform. This scalar enables validation of various array formats
    into a mikro api compliant xr.DataArray.."""

    def __init__(self, value: xr.DataArray) -> None:
        """Initialize the ArrayLike scalar with an xarray DataArray."""
        self.value = value
        self.key = str(uuid.uuid4())

    def __get__(self, instance, owner) -> "ArrayLike": ...  # noqa: ANN001, D105 #type: ignore

    def __set__(self, instance, value: ArrayCoercible) -> None: ...  # noqa: ANN001, D105 #type: ignore

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_after_validator_function(cls.validate, handler(object))

    @classmethod
    def validate(cls, v: ArrayCoercible) -> "ArrayLike":
        """Validate the input array and convert it to a xr.DataArray."""
        was_labeled = True
        # initial coercion checks, if a numpy array is passed, we need to convert it to a xarray
        # but that means the user didnt pass the dimensions explicitly so we need to add them
        # but error if they do not make sense

        if isinstance(v, np.ndarray) or is_dask_array(v):
            v = cast(NDArray[np.float_], v)  # type: ignore
            dims = ["c", "t", "z", "y", "x"]
            v = xr.DataArray(v, dims=dims[5 - v.ndim :])
            was_labeled = False

        if not isinstance(v, xr.DataArray):
            raise ValueError("This needs to be a instance of xarray.DataArray")

        if "x" not in v.dims:
            raise ValueError("Representations must always have a 'x' Dimension")

        if "y" not in v.dims:
            raise ValueError("Representations must always have a 'y' Dimension")

        if "t" not in v.dims:
            v = v.expand_dims("t")
        if "c" not in v.dims:
            v = v.expand_dims("c")
        if "z" not in v.dims:
            v = v.expand_dims("z")

        chunks = rechunk(v.sizes, itemsize=v.data.itemsize, chunksize_in_bytes=20_000_000)
        if not was_labeled:
            if v.sizes["t"] > v.sizes["x"] or v.sizes["t"] > v.sizes["y"]:
                raise ValueError(
                    f"Probably Non sensical dimensions. T is bigger than x or y: Sizes {v.sizes}"
                )
            if v.sizes["z"] > v.sizes["x"] or v.sizes["z"] > v.sizes["y"]:
                raise ValueError(
                    f"Probably Non sensical dimensions. Z is bigger than x or y: Sizes {v.sizes}"
                )
            if v.sizes["c"] > v.sizes["x"] or v.sizes["c"] > v.sizes["y"]:
                raise ValueError(
                    f"Probably Non sensical dimensions. C is bigger than x or y: Sizes {v.sizes}"
                )

        v = v.chunk({key: chunksize for key, chunksize in chunks.items() if key in v.dims})  # type: ignore

        v = v.transpose(*"ctzyx")

        if is_dask_array(v.data):
            v = v.compute()  # type: ignore

        return cls(v)

    def __repr__(self) -> str:
        """Return a string representation of the ArrayLike scalar."""
        return f"InputArray({self.value})"


class BigFile:
    """A custom scalar for wrapping of every supported array like structure on
    the mikro platform. This scalar enables validation of various array formats
    into a mikro api compliant xr.DataArray.."""

    def __init__(self, value: IO[bytes]) -> None:
        """Initialize the BigFile scalar with a file-like object."""
        self.value = value
        self.key = str(value.name)

    def __get__(self, instance, owner) -> "BigFile": ...  # noqa: ANN001, D105 # type: ignore

    def __set__(self, instance, value: FileCoercible) -> None: ...  # noqa: ANN001, D105 # type: ignore

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_after_validator_function(cls.validate, handler(object))

    @classmethod
    def validate(cls, v: FileCoercible) -> "BigFile":
        """Validate the input array and convert it to a xr.DataArray."""

        if isinstance(v, str):
            v = open(v, "rb")

        if not isinstance(v, io.IOBase):
            raise ValueError("This needs to be a instance of a file")

        return cls(v)

    def __repr__(self) -> str:
        """Return a string representation of the BigFile scalar."""
        return f"BigFile({self.value})"


class ParquetLike:
    """A custom scalar for ensuring a common format to support write to the
    parquet api supported by mikro_next It converts the passed value into
    a compliant format.."""

    def __init__(self, value: pd.DataFrame) -> None:
        """Initialize the ParquetLike scalar with a pandas DataFrame."""
        self.value = value
        self.key = str(uuid.uuid4())

    def __get__(self, instance, owner) -> "ParquetLike": ...  # noqa: ANN001, D105 # type: ignore

    def __set__(self, instance, value: ParquetCoercible) -> None: ...  # noqa: ANN001, D105 # type: ignore

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_after_validator_function(cls.validate, handler(object))

    @classmethod
    def validate(cls, v: ParquetCoercible) -> "ParquetLike":
        """Validate the validator function"""

        if not isinstance(v, pd.DataFrame):  # type: ignore
            raise ValueError("This needs to be a instance of pandas DataFrame")

        return cls(v)

    def __repr__(self) -> str:
        """Return a string representation of the ParquetLike scalar."""
        return f"ParquetLike({self.value})"


class ImageFileLike:
    """A custom scalar for ensuring a common format to support write to the
    parquet api supported by mikro_next It converts the passed value into
    a compliant format.."""

    def __init__(self, value: io.BufferedReader, name: str = "") -> None:
        """Initialize the ImageFileLike scalar with a file-like object."""
        self.value = value
        self.file_name = os.path.basename(name)
        self.mime_type = mimetypes.guess_type(self.file_name)[0]

    def __get__(self, instance, owner) -> "FileLike": ...  # noqa: ANN001, D105 # type: ignore

    def __set__(self, instance, value: FileCoercible) -> None: ...  # noqa: ANN001, D105 # type: ignore

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_after_validator_function(cls.validate, handler(object))

    @classmethod
    def validate(cls, v: FileCoercible) -> "ImageFileLike":
        """Validate the validator function"""

        if isinstance(v, str):
            file = open(v, "rb")
            name = v
        elif isinstance(v, io.IOBase):
            file = v
            name = v.name
        elif isinstance(v, Path):
            file = open(v, "rb")
            name = str(v)
        else:
            raise ValueError(
                f"Unsupported type {type(v)}. Please provide a string or a Path object. Or a file object that is opened in binary mode."
            )

        if not isinstance(file, io.BufferedReader):  # type: ignore
            raise ValueError("This needs to be a instance of a file")

        return cls(file, name=name)

    def __repr__(self) -> str:
        """Return a string representation of the ImageFileLike scalar."""
        return f"FileLike({self.value})"


class FileLike:
    """A custom scalar for ensuring a common format to support write to the
    parquet api supported by mikro_next It converts the passed value into
    a compliant format.."""

    def __init__(self, value: IO[bytes], name: str = "") -> None:
        """Initialize the FileLike scalar with a file-like object."""
        self.value = value
        self.file_name = os.path.basename(name)
        self.mime_type = mimetypes.guess_type(self.file_name)[0]

    def __get__(self, instance, owner) -> "FileLike": ...  # noqa: ANN001, D105 # type: ignore

    def __set__(self, instance, value: FileCoercible) -> None: ...  # noqa: ANN001, D105 # type: ignore

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_after_validator_function(cls.validate, handler(object))

    @classmethod
    def validate(cls, v: FileCoercible) -> "FileLike":
        """Validate the validator function"""

        if isinstance(v, str):
            file = open(v, "rb")
            name = v
        elif isinstance(v, io.IOBase):
            file = v
            name = v.name
        elif isinstance(v, Path):
            file = open(v, "rb")
            name = str(v)
        else:
            raise ValueError(
                f"Unsupported type {type(v)}. Please provide a string or a Path object. Or a file object that is opened in binary mode."
            )

        if not isinstance(file, io.IOBase):  # type: ignore
            raise ValueError("This needs to be a instance of a file")

        return cls(file, name=name)

    def __repr__(self) -> str:
        """Return a string representation of the FileLike scalar."""
        return f"FileLike({self.value})"


class MeshLike:
    """A custom scalar for ensuring a common format to support write to the
    mesh api supported by mikro_next It converts the passed value into
    a compliant format.."""

    def __init__(self, value: IO[bytes], name: str = "") -> None:
        """Initialize the MeshLike scalar with a file-like object."""
        self.value = value
        self.key = str(name)

    def __get__(self, instance, owner) -> "MeshLike":  # noqa: ANN001 # type: ignore
        """Get the MeshLike scalar from the instance."""
        ...

    def __set__(self, instance, value: MeshCoercible):  # noqa: ANN001, ANN204 # type: ignore
        """Set the MeshLike scalar on the instance."""
        ...  # noqa: ANN001

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,  # noqa: ANN401
    ) -> CoreSchema:
        """Get the pydantic core schema for the validator function"""
        return core_schema.no_info_after_validator_function(cls.validate, handler(object))

    @classmethod
    def validate(cls, v: MeshCoercible) -> "MeshLike":
        """Validate the input array and convert it to a xr.DataArray."""

        if isinstance(v, str):
            file = open(v, "rb")
            name = v
        elif isinstance(v, io.IOBase):
            file = v
            name = "Random Name"
        else:
            file = v
            name = str(v)

        if not isinstance(file, io.BufferedReader):
            raise ValueError("This needs to be a instance of a file")

        return cls(file, name=name)

    def __repr__(self) -> str:
        """Return a string representation of the MeshLike scalar."""
        return f"MeshLike({self.value})"



from typing import Any, Dict

from fakts_next.fakts import Fakts
from mikro_next.datalayer import DataLayer
from pydantic import BaseModel


class DataLayerFakt(BaseModel):
    endpoint_url: str


class FaktsDataLayer(DataLayer):
    """A fakts implementation of the datalayer. This will allow you to connect to a datalayer
    that is defined asnychronously in fakts. This is useful for connecting to a datalayer that
    is not known at compile time. Will get the server configuration from fakts and connect to the
    datalayer."""

    fakts_group: str
    fakts: Fakts

    _old_fakt: Dict[str, Any] = {}
    _configured = False

    async def get_endpoint_url(self) -> str:
        """Get the endpoint URL for the datalayer. This will connect to the fakts group and get the
        endpoint URL from the fakts group. If the datalayer is already configured, it will return the
        endpoint URL. If not, it will connect to the fakts group and get the endpoint URL."""
        if self._configured:
            return self.endpoint_url
        else:
            await self.aconnect()
            return self.endpoint_url

    async def aconnect(self) -> None:
        """Connect to the fakts group and get the endpoint URL. This will set the endpoint URL to the
        fakts group alias. If the fakts group is not configured, it will raise an error."""
        alias = await self.fakts.aget_alias(self.fakts_group)
        self.endpoint_url = alias.to_http_path()

        self._configured = True



from pydantic import Field
from koil.composition import Composition
from mikro_next.rath import MikroNextRath
from mikro_next.datalayer import DataLayer


class MikroNext(Composition):
    """The Mikro Composition

    This composition provides a datalayer and a rath for interacting with the
    mikro api and beyond

    You shouldn't need to create this directly, instead use the builder functions
    to generate a new instance of this composition.

    ```python

    from mikro import Mikro

    async def aget_token():
        return "XXXX"

    m = Mikro(
        datalayer= DataLayer(endpoint_url="s3.amazonaws.com", access_key="XXXX", secret_key="XXXX"),
        mikro = MikroRath(link=MikroLinkComposition(auth=AuthTokenLink(token_loader=aget_token)))),
    )
    ```
    """

    datalayer: DataLayer = Field(
        ..., description="The datalayer for interacting with the mikro api"
    )
    rath: MikroNextRath



"""Functions for executing queries and mutations using Rekuest Next and Rath.

This function hook into the api funciton that are autogenerated by turms and
allow you to execute queries and mutations using thre rekuest-rath client.

"""

from typing import Any, Dict, Generator, AsyncGenerator, Type
from mikro_next.rath import MikroNextRath, current_mikro_next_rath
from koil import unkoil, unkoil_gen
from rath.turms.funcs import TOperation
from .errors import NoMikroFound


def execute(
    operation: Type[TOperation],
    variables: Dict[str, Any],
    rath: MikroNextRath | None = None,
) -> TOperation:
    """Executes a query or mutation using rath in a blocking way."""
    return unkoil(aexecute, operation, variables, rath)


async def aexecute(
    operation: Type[TOperation],
    variables: Dict[str, Any],
    rath: MikroNextRath | None = None,
) -> TOperation:
    """Executes a query or mutation using rath in a non-blocking way."""
    rath = rath or current_mikro_next_rath.get()
    if not rath:
        raise NoMikroFound(
            "No rath client found in context. Please provide a rath client."
        )

    x = await rath.aquery(
        operation.Meta.document,
        operation.Arguments(**variables).model_dump(by_alias=True, exclude_unset=True),
    )
    return operation(**x.data)


def subscribe(
    operation: Type[TOperation],
    variables: Dict[str, Any],
    rath: MikroNextRath | None = None,
) -> Generator[TOperation, None, None]:
    """Subscribes to a query or mutation using rath in a blocking way."""
    return unkoil_gen(asubscribe, operation, variables, rath)


async def asubscribe(
    operation: Type[TOperation],
    variables: Dict[str, Any],
    rath: MikroNextRath | None = None,
) -> AsyncGenerator[TOperation, None]:
    """Subscribes to a query or mutation using rath in a non-blocking way."""
    rath = rath or current_mikro_next_rath.get()
    if not rath:
        raise NoMikroFound(
            "No rath client found in context. Please provide a rath client."
        )

    async for event in rath.asubscribe(
        operation.Meta.document,
        operation.Arguments(**variables).model_dump(by_alias=True),
    ):
        yield operation(**event.data)



"""Module for uploading various data types to a DataLayer using asynchronous methods."""

import os
from typing import TYPE_CHECKING
from mikro_next.scalars import ArrayLike, ImageFileLike, MeshLike, ParquetLike, FileLike
import asyncio
import s3fs  # type: ignore
from aiobotocore.session import get_session  # type: ignore
import botocore  # type: ignore
from concurrent.futures import ThreadPoolExecutor

from .errors import PermissionsError, UploadError
from zarr.storage import FsspecStore
import zarr.api.asynchronous as async_api
import aiohttp

if TYPE_CHECKING:
    from mikro_next.api.schema import Credentials, PresignedPostCredentials
    from mikro_next.datalayer import DataLayer


async def astore_xarray_input(
    xarray: ArrayLike,
    credentials: "Credentials",
    endpoint_url: str,
) -> str:
    """Stores an xarray in the DataLayer"""

    if endpoint_url.startswith("https://"):
        os.environ["AWS_REQUEST_CHECKSUM_CALCULATION"] = (
            "when_required"  # TODO: This is a workaround for a bug in aiobotocore and s3fs https://github.com/fsspec/s3fs/issues/931
        )

    filesystem = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=True,
    )

    # random_uuid = uuid.uuid4()
    # s3_path = f"zarr/{random_uuid}.zarr"

    array = xarray.value.transpose("c", "t", "z", "y", "x")

    s3_path = f"{credentials.bucket}/{credentials.key}"
    store = FsspecStore(filesystem, read_only=False, path=s3_path)

    try:
        await async_api.save_array(store, array.to_numpy(), zarr_format=3)  # type: ignore
        return credentials.store
    except Exception as e:
        raise UploadError(f"Error while uploading to {s3_path} on {endpoint_url}") from e


def _store_parquet_input(
    parquet_input: ParquetLike,
    credentials: "Credentials",
    endpoint_url: str,
) -> str:
    """Stores an xarray in the DataLayer"""
    import pyarrow.parquet as pq  # type: ignore
    from pyarrow import Table  # type: ignore

    filesystem = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
    )

    table: Table = Table.from_pandas(parquet_input.value)  # type: ignore

    s3_path = f"s3://{credentials.bucket}/{credentials.key}"
    try:
        pq.write_table(table, s3_path, filesystem=filesystem)  # type: ignore
        return credentials.store
    except Exception as e:
        raise UploadError(f"Error while uploading to {s3_path}") from e


async def astore_mesh_file(
    mesh: MeshLike,
    credentials: "PresignedPostCredentials",
    datalayer: "DataLayer",
) -> str:
    """Store a mesh file in the DataLayer using presigned POST credentials."""
    endpoint_url = await datalayer.get_endpoint_url()

    async with aiohttp.ClientSession() as session:
        form_data = aiohttp.FormData()
        form_data.add_field("key", credentials.key)
        form_data.add_field("policy", credentials.policy)
        form_data.add_field("x-amz-algorithm", credentials.x_amz_algorithm)
        form_data.add_field("x-amz-credential", credentials.x_amz_credential)
        form_data.add_field("x-amz-date", credentials.x_amz_date)
        form_data.add_field("x-amz-signature", credentials.x_amz_signature)
        form_data.add_field(
            "file",
            mesh.value,
            filename="mesh.obj",
            content_type="application/octet-stream",
        )

        url = endpoint_url + "/" + credentials.bucket

        async with session.post(url, data=form_data) as resp:
            if resp.status not in {200, 204}:
                body = await resp.text()
                raise UploadError(f"Error while uploading mesh: HTTP {resp.status}: {body}")

    return credentials.store


async def astore_media_file(
    file: ImageFileLike,
    credentials: "PresignedPostCredentials",
    datalayer: "DataLayer",
) -> str:
    """Store a media file in the DataLayer using presigned POST credentials."""
    endpoint_url = await datalayer.get_endpoint_url()

    async with aiohttp.ClientSession() as session:
        form_data = aiohttp.FormData()
        form_data.add_field("key", credentials.key)
        form_data.add_field("policy", credentials.policy)
        form_data.add_field("x-amz-algorithm", credentials.x_amz_algorithm)
        form_data.add_field("x-amz-credential", credentials.x_amz_credential)
        form_data.add_field("x-amz-date", credentials.x_amz_date)
        form_data.add_field("x-amz-signature", credentials.x_amz_signature)
        form_data.add_field(
            "file",
            file.value,
            filename=file.file_name,
            content_type="application/octet-stream",
        )

        url = endpoint_url + "/" + credentials.bucket

        async with session.post(url, data=form_data) as resp:
            if resp.status not in {200, 204}:
                body = await resp.text()
                raise UploadError(f"Error while uploading mesh: HTTP {resp.status}: {body}")

    print("Successfully uploaded media file", credentials)
    return credentials.store


async def aupload_bigfile(
    file: FileLike | ImageFileLike,
    credentials: "Credentials",
    datalayer: "DataLayer",
) -> str:
    """Store a DataFrame in the DataLayer"""
    session = get_session()

    endpoint_url = await datalayer.get_endpoint_url()

    async with session.create_client(  # type: ignore
        "s3",
        region_name="us-west-2",
        endpoint_url=endpoint_url,
        aws_secret_access_key=credentials.secret_key,
        aws_access_key_id=credentials.access_key,
        aws_session_token=credentials.session_token,
    ) as client:
        try:
            print(credentials, file.value)
            await client.put_object(Bucket=credentials.bucket, Key=credentials.key, Body=file.value)  # type: ignore
        except botocore.exceptions.ClientError as e:  # type: ignore
            if e.response["Error"]["Code"] == "InvalidAccessKeyId":  # type: ignore
                return PermissionsError("Access Key is invalid, trying to get new credentials")  # type: ignore

            raise e

    print(credentials)
    return credentials.store


async def aupload_xarray(
    array: ArrayLike,
    credentials: "Credentials",
    datalayer: "DataLayer",
) -> str:
    """Store a DataFrame in the DataLayer"""
    return await astore_xarray_input(array, credentials, await datalayer.get_endpoint_url())


async def aupload_parquet(
    parquet: ParquetLike,
    credentials: "Credentials",
    datalayer: "DataLayer",
    executor: ThreadPoolExecutor,
) -> str:
    """Store a DataFrame in the DataLayer"""
    co_future = executor.submit(
        _store_parquet_input, parquet, credentials, await datalayer.get_endpoint_url()
    )
    return await asyncio.wrap_future(co_future)



from .mikro_next import MikroNext
from .utils import rechunk

try:
    #
    from .arkitekt import MikroService
except ImportError as e:
    try:
        import arkitekt

        raise ImportError(
            "Arkitekt is installed, but the MikroService could not be imported. This may indicate a version mismatch or missing dependencies."
        ) from e
    except ImportError:
        pass


try:
    from .rekuest import structure_reg

    print("Imported structure_reg")
except ImportError as e:
    print("Could not import structure_reg", e)
    pass


__all__ = [
    "MikroNext",
    "rechunk",
    "structure_reg",
    "MikroService",
]



import zarr.storage
from mikro_next.api.schema import (
    arequest_access,
    arequest_table_access,
    arequest_file_access,
    AccessCredentials,
)
from mikro_next.datalayer import DataLayer, current_next_datalayer
import s3fs  # type: ignore
from koil import unkoil
import zarr
import aiohttp
from typing import Tuple
from typing import TYPE_CHECKING

from rath.scalars import ID

if TYPE_CHECKING:
    import pyarrow.parquet as pq  # type: ignore


async def aget_zarr_credentials_and_endpoint(
    store: str,
) -> Tuple[AccessCredentials, str]:
    datalayer = current_next_datalayer.get()
    if not datalayer:
        raise ValueError("Datalayer is not set")
    credentials = await arequest_access(ID.validate(store))

    endpoint_url = await datalayer.get_endpoint_url()
    return credentials, endpoint_url


async def aget_table_credentials_and_endpoint(
    store: str,
) -> Tuple[AccessCredentials, str]:
    datalayer = current_next_datalayer.get()
    if not datalayer:
        raise ValueError("Datalayer is not set")

    credentials = await arequest_table_access(ID.validate(store))
    endpoint_url = await datalayer.get_endpoint_url()
    return credentials, endpoint_url


async def aget_file_credentials_and_endpoint(
    store: str,
) -> Tuple[AccessCredentials, str]:
    datalayer = current_next_datalayer.get()
    if not datalayer:
        raise ValueError("Datalayer is not set")

    credentials = await arequest_file_access(ID.validate(store))
    endpoint_url = await datalayer.get_endpoint_url()
    return credentials, endpoint_url


async def aopen_zarr_store(store_id: str, cache: int = 2**30):
    credentials, endpoint_url = await aget_zarr_credentials_and_endpoint(store_id)

    _s3fs = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=True,
    )

    return zarr.storage.FsspecStore(
        _s3fs, read_only=False, path=f"{credentials.bucket}/{credentials.key}"
    )


def open_zarr_store(store_id: str, cache: int = 2**30):
    credentials, endpoint_url = unkoil(aget_zarr_credentials_and_endpoint, store_id)

    _s3fs = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=True,
    )
    return zarr.storage.FsspecStore(
        _s3fs, read_only=False, path=f"{credentials.bucket}/{credentials.key}"
    )


async def aopen_parquet_filesytem(store_id: str):
    try:
        import pyarrow.parquet as pq  # type: ignore
    except ImportError as e:
        raise ImportError("You need to install pyarrow to use this function") from e
    credentials, endpoint_url = await aget_table_credentials_and_endpoint(store_id)

    _s3fs = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=True,
    )

    return pq.ParquetDataset(credentials.path, filesystem=_s3fs)


def open_parquet_filesystem(store_id: str):
    credentials, endpoint_url = unkoil(aget_table_credentials_and_endpoint, store_id)

    _s3fs = s3fs.S3FileSystem(
        secret=credentials.secret_key,
        key=credentials.access_key,
        client_kwargs={
            "endpoint_url": endpoint_url,
            "aws_session_token": credentials.session_token,
        },
        asynchronous=False,
    )

    return pq.ParquetDataset(credentials.path, filesystem=_s3fs)


async def adownload_file(
    presigned_url: str,
    file_name: str,
    datalayer: DataLayer | None = None,
):
    datalayer = datalayer or current_next_datalayer.get()
    if not datalayer:
        raise ValueError("Datalayer is not set")
    endpoint_url = await datalayer.get_endpoint_url()

    async with aiohttp.ClientSession() as session:
        async with session.get(endpoint_url + presigned_url) as response:
            with open(file_name, "wb") as file:
                while True:
                    chunk = await response.content.read(
                        1024
                    )  # read the response by chunks of 1024 bytes
                    if not chunk:
                        break
                    file.write(chunk)

    return file_name


def download_file(
    presigned_url: str, file_name: str, datalayer: DataLayer | None = None
):
    return unkoil(
        adownload_file, presigned_url, file_name=file_name, datalayer=datalayer
    )



class MikroError(Exception):
    """Base class for all Mikro errors."""


class NoMikroFound(MikroError):
    """Caused when no Mikro is found."""


class NoDataLayerFound(MikroError):
    """Caused when no DataLayer is found."""


class NotQueriedError(MikroError):
    """Caused when a field has not been queries"""



from rekuest_next.structures.default import (
    get_default_structure_registry,
    id_shrink,
)
from rekuest_next.widgets import SearchWidget
from mikro_next.api.schema import (
    Image,
    Snapshot,
    ROI,
    Stage,
    Dataset,
    File,
    RGBContext,
    Mesh,
    Table,
    TableCell,
    TableRow,
    RGBView,
    aget_rgb_view,
    aget_image,
    aget_snapshot,
    aget_roi,
    aget_stage,
    aget_dataset,
    aget_file,
    aget_rgb_context,
    aget_mesh,
    aget_table,
    aget_table_cell,
    aget_table_row,
    SearchRGBViewsQuery,
    SearchImagesQuery,
    SearchSnapshotsQuery,
    SearchRoisQuery,
    SearchStagesQuery,
    SearchDatasetsQuery,
    SearchFilesQuery,
    SearchMeshesQuery,
    SearchTablesQuery,
    SearchTableCellsQuery,
    SearchTableRowsQuery,
)


structure_reg = get_default_structure_registry()

structure_reg.register_as_structure(
    Image,
    identifier="@mikro/image",
    aexpand=aget_image,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchImagesQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    Snapshot,
    identifier="@mikro/snapshot",
    aexpand=aget_snapshot,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchSnapshotsQuery.Meta.document, ward="mikro"),
)

structure_reg.register_as_structure(
    ROI,
    identifier="@mikro/roi",
    aexpand=aget_roi,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchRoisQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    Stage,
    identifier="@mikro/stage",
    aexpand=aget_stage,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchStagesQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    RGBView,
    identifier="@mikro/rgbview",
    aexpand=aget_rgb_view,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchRGBViewsQuery.Meta.document, ward="mikro"),
)


structure_reg.register_as_structure(
    Stage,
    identifier="@mikro/stage",
    aexpand=aget_stage,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchStagesQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    Dataset,
    identifier="@mikro/dataset",
    aexpand=aget_dataset,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchDatasetsQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    File,
    identifier="@mikro/file",
    aexpand=aget_file,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchFilesQuery.Meta.document, ward="mikro"),
)
structure_reg.register_as_structure(
    RGBContext,
    identifier="@mikro/rbgcontext",
    aexpand=aget_rgb_context,
    ashrink=id_shrink,
)


structure_reg.register_as_structure(
    Mesh,
    identifier="@mikro/mesh",
    aexpand=aget_mesh,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchMeshesQuery.Meta.document, ward="mikro"),
)

structure_reg.register_as_structure(
    Table,
    identifier="@mikro/table",
    aexpand=aget_table,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchTablesQuery.Meta.document, ward="mikro"),
)

structure_reg.register_as_structure(
    TableCell,
    identifier="@mikro/tablecell",
    aexpand=aget_table_cell,
    ashrink=id_shrink,
    default_widget=SearchWidget(
        query=SearchTableCellsQuery.Meta.document, ward="mikro"
    ),
)

structure_reg.register_as_structure(
    TableRow,
    identifier="@mikro/tablerow",
    aexpand=aget_table_row,
    ashrink=id_shrink,
    default_widget=SearchWidget(query=SearchTableRowsQuery.Meta.document, ward="mikro"),
)



from types import TracebackType
from pydantic import Field
from .links.upload import UploadLink
from rath import rath
import contextvars
from rath.links.auth import AuthTokenLink
from rath.links.compose import TypedComposedLink
from rath.links.dictinglink import DictingLink
from rath.links.file import FileExtraction
from rath.links.split import SplitLink
from typing import Optional


current_mikro_next_rath: contextvars.ContextVar[Optional["MikroNextRath"]] = (
    contextvars.ContextVar("current_mikro_next_rath")
)


class MikroNextLinkComposition(TypedComposedLink):
    """The MikroNextLinkComposition

    This is a composition of links that are traversed before a request is sent to the
    mikro api. This link composition contains the default links for mikro_next.

    You shouldn't need to create this directly.
    """

    fileextraction: FileExtraction = Field(default_factory=FileExtraction)
    """ A link that extracts files from the request and follows the graphql multipart request spec"""
    dicting: DictingLink = Field(default_factory=DictingLink)
    """ A link that converts basemodels to dicts"""
    upload: UploadLink
    """ A link that uploads supported data types like numpy arrays and parquet files to the datalayer"""
    auth: AuthTokenLink
    """ A link that splits the request into a http and a websocket request"""
    split: SplitLink


class MikroNextRath(rath.Rath):
    """Mikro Rath

    Mikro Rath is the GraphQL client for mikro_next It is a thin wrapper around Rath
    that provides some default links and a context manager to set the current
    client. (This allows you to use the `mikro_nextrath.current` function to get the
    current client, within the context of mikro app).

    This is a subclass of Rath that adds some default links to convert files and array to support
    the graphql multipart request spec."""

    async def __aenter__(self) -> "MikroNextRath":
        """Sets the current mikro_next rath to this instance"""
        await super().__aenter__()
        current_mikro_next_rath.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Resets the current mikro_next rath to None"""
        await super().__aexit__(exc_type, exc_val, exc_tb)
        current_mikro_next_rath.set(None)



import json
import os
from typing import Any, Dict
from fakts_next.contrib.rath.auth import FaktsAuthLink
from rath.links.dictinglink import DictingLink
from rath.links.file import FileExtraction
from rath.links.split import SplitLink
from fakts_next import Fakts
from arkitekt_next.service_registry import BaseArkitektService, Params
from fakts_next.models import Requirement

from mikro_next.mikro_next import MikroNext
from mikro_next.rath import MikroNextRath
from rath.links.compose import compose
from fakts_next.contrib.rath.aiohttp import FaktsAIOHttpLink
from fakts_next.contrib.rath.graphql_ws import FaktsGraphQLWSLink
from mikro_next.contrib.fakts.datalayer import FaktsDataLayer
from mikro_next.links.upload import UploadLink
from graphql import OperationType
from arkitekt_next.service_registry import (
    get_default_service_registry,
)

from rekuest_next.links.context import ContextLink


def build_relative_path(*path: str) -> str:
    """Builds a relative path to the given path components, starting from the
    directory of this file. This is useful for building paths to files that are"""
    return os.path.join(os.path.dirname(__file__), *path)


class MikroService(BaseArkitektService):
    """A service that allows you to connect to a MikroNext instance within your
    Arkitekt Application. This service will allow you to make requests to the"""

    def get_service_name(self) -> str:
        """Returns the name of the service."""
        return "mikro"

    def build_service(self, fakts: Fakts, params: Params) -> MikroNext:
        """Builds the MikroNext service with the given parameters.
        Args:
            fakts (Fakts): The Fakts instance to use for the service.
            params (Params): The parameters for the service.
        Returns:
            MikroNext: An instance of MikroNext with the configured links.
        """
        datalayer = FaktsDataLayer(fakts_group="datalayer", fakts=fakts)

        return MikroNext(
            rath=MikroNextRath(
                link=compose(
                    FileExtraction(),
                    DictingLink(),
                    ContextLink(),
                    UploadLink(
                        datalayer=datalayer,
                    ),
                    FaktsAuthLink(
                        fakts=fakts,
                    ),  # needs to be after the UploadLink as the upload link will also use the auth link
                    SplitLink(
                        left=FaktsAIOHttpLink(
                            fakts_group="mikro",
                            fakts=fakts,
                            endpoint_url="FAKE_URL",
                        ),
                        right=FaktsGraphQLWSLink(
                            fakts_group="mikro",
                            fakts=fakts,
                            ws_endpoint_url="FAKE_URL",
                        ),
                        split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                    ),
                ),
            ),
            datalayer=datalayer,
        )

    def get_requirements(self) -> list[Requirement]:
        """Returns a list of requirements for the Mikro service."""
        return [
            Requirement(
                key="mikro",
                service="live.arkitekt.mikro",
                description="An instance of ArkitektNext Mikro to make requests to the user's data",
                optional=True,
            ),
            Requirement(
                key="datalayer",
                service="live.arkitekt.s3",
                description="An instance of ArkitektNext Datalayer to make requests to the user's data",
                optional=True,
            ),
        ]

    def get_graphql_schema(self) -> str:
        """Returns the GraphQL schema for the Mikro service."""
        schema_graphql_path = build_relative_path("api", "schema.graphql")
        with open(schema_graphql_path) as f:
            return f.read()

    def get_turms_project(self) -> Dict[str, Any]:
        """Returns the Turms project configuration for the Mikro service.

        This will be used to generate the Turms project configuration when
        autogenerating code for the graphql queries and mutations for the Mikro service.

        """
        turms_prject = build_relative_path("api", "project.json")
        with open(turms_prject) as f:
            return json.loads(f.read())


# Register the MikroService with the default service registry
get_default_service_registry().register(MikroService())



"""Default Widgets

This Module provides default widgets that can be used with the mikro app in an arkitekt
context.

Attributes:
    MY_TOP_REPRESENTATIONS (SearchWidget): The top representations for the currently active user
    MY_TOP_SAMPLES (SearchWidget): The top samples for the currently active user
"""



"""
Traits for mikro_next

Traits are mixins that are added to every graphql type that exists on the mikro schema.
We use them to add functionality to the graphql types that extend from the base type.

Every GraphQL Model on Mikro gets a identifier and shrinking methods to ensure the compatibliity
with arkitekt. This is done by adding the identifier and the shrinking methods to the graphql type.
If you want to add your own traits to the graphql type, you can do so by adding them in the graphql
.config.yaml file.

"""

from typing import Awaitable, List, Type, TypeVar, Tuple, Protocol, Optional
import numpy as np
from numpy.typing import NDArray
from pydantic import BaseModel
import xarray as xr
import pandas as pd
from typing import TYPE_CHECKING
from dask.array.core import from_zarr  # type: ignore
from zarr.storage import FsspecStore
from .scalars import FiveDVector
from rath.scalars import ID
from typing import Any
from rath.turms.utils import get_attributes_or_error
from rath.traits import FederationFetchable
from rath.rath import Rath

TwoDArray = NDArray[np.generic]
OneDArray = NDArray[np.generic]


if TYPE_CHECKING:
    from pyarrow.parquet import ParquetDataset  # type: ignore
    from mikro_next.api.schema import HasZarrStoreAccessor


class MikroFetchable(FederationFetchable):
    """A trait for Mikro Fetchable objects

    This trait allows to fetch an object from the mikro service using its ID.
    It is used to ensure that the object can be fetched from the mikro service.

    """

    @classmethod
    def get_rath(cls) -> "Rath":
        """Get the current Rath client from the context."""
        from mikro_next.rath import current_mikro_next_rath

        return current_mikro_next_rath.get()


class HasZarrStoreTrait(BaseModel):
    """Image Trait

    Implements both identifier and shrinking methods.
    Also Implements the data attribute

    Attributes:
        data (xarray.Dataset): The data of the representation.

    """

    @property
    def data(self) -> xr.DataArray:
        """The data of this image as an xarray.DataArray"""
        from dask.array.core import Array

        store: HasZarrStoreAccessor = get_attributes_or_error(self, "store")

        array: Array = from_zarr(store.zarr_store)

        return xr.DataArray(array, dims=["c", "t", "z", "y", "x"])

    @property
    def multi_scale_data(self) -> List[xr.DataArray]:
        """The multi-scale data of this image as a list of xarray.DataArray"""
        scale_views = get_attributes_or_error(self, "derived_scale_views")

        if len(scale_views) == 0:
            raise ValueError(
                "No ScaleView found in views. Please create a ScaleView first."
            )

        sorted_views = sorted(scale_views, key=lambda image: image.scale_x)
        return [x.image.data for x in sorted_views]

    async def adata(self) -> Awaitable[xr.DataArray]:
        """The Data of the Representation as an xr.DataArray. Accessible from asyncio.

        Returns:
            xr.DataArray: The associated object.

        Raises:
            AssertionError: If the representation has no store attribute quries
        """
        pstore = get_attributes_or_error(self, "store")
        return await pstore.aopen()

    def get_pixel_size(self, stage: ID | None = None) -> Tuple[float, float, float]:
        """The pixel size of the representation

        Returns:
            Tuple[float, float, float]: The pixel size
        """
        from mikro_next.api.schema import AffineTransformationView

        views = get_attributes_or_error(self, "views")

        for view in views:
            if isinstance(view, AffineTransformationView):
                if stage is None:
                    return view.affine_matrix.as_matrix().diagonal()[:3]  # type: ignore
                else:
                    if get_attributes_or_error(view, "stage.id") == stage:
                        return view.affine_matrix.as_matrix().diagonal()[:3]  # type: ignore

        raise NotImplementedError(
            f"No pixel size found for this representation {self}. Have you attached any views?"
        )


class PhysicalSizeProtocol(Protocol):
    """A Protocol for Vectorizable data

    Attributes:
        x (float): The x value
        y (float): The y value
        z (float): The z value
        t (float): The t value
        c (float): The c value
    """

    x: float | None
    y: float | None
    z: float | None
    t: float | None
    c: float | None

    def __init__(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        t: Optional[float] = None,
        c: Optional[float] = None,
    ) -> None:
        """Initialize the PhysicalSizeProtocol."""
        ...


class PhysicalSizeTrait:
    """Additional Methods for PhysicalSize"""

    def is_similar(
        self: PhysicalSizeProtocol,
        other: PhysicalSizeProtocol,
        tolerance: float = 0.02,
        raise_exception: Optional[bool] = False,
    ) -> bool:
        if hasattr(self, "x") and self.x is not None and other.x is not None:
            if abs(other.x - self.x) > tolerance:
                if raise_exception:
                    raise ValueError(
                        f"X values are not similar: {self.x} vs {other.x} is above tolerance {tolerance}"
                    )
                return False
        if hasattr(self, "y") and self.y is not None and other.y is not None:
            if abs(other.y - self.y) > tolerance:
                if raise_exception:
                    raise ValueError(
                        f"Y values are not similar: {self.y} vs {other.y} is above tolerance {tolerance}"
                    )
                return False
        if hasattr(self, "z") and self.z is not None and other.z is not None:
            if abs(other.z - self.z) > tolerance:
                if raise_exception:
                    raise ValueError(
                        f"Z values are not similar: {self.z} vs {other.z} is above tolerance {tolerance}"
                    )
                return False
        if hasattr(self, "t") and self.t is not None and other.t is not None:
            if abs(other.t - self.t) > tolerance:
                if raise_exception:
                    raise ValueError(
                        f"T values are not similar: {self.t} vs {other.t} is above tolerance {tolerance}"
                    )
                return False
        if hasattr(self, "c") and self.c is not None and other.c is not None:
            if abs(other.c - self.c) > tolerance:
                if raise_exception:
                    raise ValueError(
                        f"C values are not similar: {self.c} vs {other.c} is above tolerance {tolerance}"
                    )
                return False

        return True

    def to_scale(self) -> List[float]:
        """Get the scale of the physical size
        Returns:
            List[float]: The scale of the physical size
        """
        return [
            getattr(self, "t", 1),
            getattr(self, "c", 1),
            getattr(self, "z", 1),
            getattr(self, "y", 1),
            getattr(self, "x", 1),
        ]


class IsVectorizableTrait:
    """Additional Methods for ROI"""

    @property
    def vector_data(self) -> TwoDArray:
        """A numpy array of the vectors of the ROI

        Returns:
            np.ndarray: _description_
        """
        return self.get_vector_data(dims="yx")

    def get_vector_data(self, dims: str | list[str] = "yx") -> TwoDArray:
        """Get the vector data of the ROI as a numpy array"""
        vector_list = get_attributes_or_error(self, "vectors")
        assert vector_list, (
            "Please query 'vectors' in your request on 'ROI'. Data is not accessible otherwise"
        )
        vector_list: list[list[float]]

        mapper = {
            "y": 4,
            "x": 3,
            "z": 2,
            "t": 1,
            "c": 0,
        }

        return np.array([[v[mapper[ac]] for ac in dims] for v in vector_list])

    def center(self) -> FiveDVector:
        """The center of the ROI

        Caluclates the geometrical center of the ROI according to its type
        and the vectors of the ROI.

        Returns:
            InputVector: The center of the ROI
        """
        from mikro_next.api.schema import RoiKind, FiveDVector

        kind = get_attributes_or_error(self, "kind")
        if kind == RoiKind.RECTANGLE:
            return FiveDVector.validate(self.get_vector_data(dims="ctzyx").mean(axis=0))

        raise NotImplementedError(
            f"Center calculation not implemented for this ROI type {kind}"
        )

    def crop(self, data: xr.DataArray) -> xr.DataArray:
        """Crop the data to the ROI

        Args:
            data (xr.DataArray): The data to crop

        Returns:
            xr.DataArray: The cropped data
        """
        vector_data = self.get_vector_data(dims="ctzyx")
        return data.sel(  # type: ignore
            x=slice(vector_data[:, 3].min(), vector_data[:, 3].max()),
            y=slice(vector_data[:, 4].min(), vector_data[:, 4].max()),
            z=slice(vector_data[:, 2].min(), vector_data[:, 2].max()),
            t=slice(vector_data[:, 1].min(), vector_data[:, 1].max()),
            c=slice(vector_data[:, 0].min(), vector_data[:, 0].max()),
        )  # type: ignore

    def center_as_array(self) -> OneDArray:
        """The center of the ROI

        Caluclates the geometrical center of the ROI according to its type
        and the vectors of the ROI.

        Returns:
            InputVector: The center of the ROI
        """
        from mikro_next.api.schema import RoiKind

        kind = get_attributes_or_error(self, "kind")
        if kind == RoiKind.RECTANGLE:
            return self.get_vector_data(dims="ctzyx").mean(axis=0)
        if kind == RoiKind.POINT:
            return self.get_vector_data(dims="ctzyx")[0]

        raise NotImplementedError(
            f"Center calculation not implemented for this ROI kind {kind}"
        )


class HasParquestStoreTrait(BaseModel):
    """Table Trait

    Implements both identifier and shrinking methods.
    Also Implements the data attribute

    Attributes:
        data (pd.DataFrame): The data of the table.

    """

    @property
    def data(self) -> pd.DataFrame:
        """The data of this table as a pandas dataframe

        Returns:
            pd.DataFrame: The Dataframe
        """
        store: "HasParquetStoreAccesor" = get_attributes_or_error(self, "store")
        return store.parquet_dataset.read_pandas().to_pandas()  # type: ignore


V = TypeVar("V")


class HasZarrStoreAccessor(BaseModel):
    """Zarr Store Accessor

    Allows to access the python zarr store of
    a ZarrStore object.


    """

    _openstore: Any = None

    @property
    def zarr_store(self) -> FsspecStore:
        """The zarr store of the ZarrStore object"""
        from mikro_next.io.download import open_zarr_store

        if self._openstore is None:
            id = get_attributes_or_error(self, "id")
            self._openstore = open_zarr_store(id)
        return self._openstore


class HasParquetStoreAccesor(BaseModel):
    """Parquet Store Accessor"""

    _dataset: Any = None

    @property
    def parquet_dataset(self) -> "ParquetDataset":
        """The Parquet Dataset of the ParquetStore object"""
        from mikro_next.io.download import open_parquet_filesystem

        if self._dataset is None:
            id = get_attributes_or_error(self, "id")
            self._dataset = open_parquet_filesystem(id)
        return self._dataset


class HasDownloadAccessor(BaseModel):
    """Download Accessor"""

    _dataset: Any = None

    def download(self, file_name: str | None = None) -> "str":
        """Download the file from the presigned URL

        Args:
            file_name (str | None): The name of the file to save the downloaded file as
                If None, the key from the presigned URL will be used as the file name.
        Returns:
            str: The path to the downloaded file
        """
        from mikro_next.io.download import download_file

        url, key = get_attributes_or_error(self, "presigned_url", "key")
        return download_file(url, file_name=file_name or key)


class HasPresignedDownloadAccessor(BaseModel):
    """Presigned Download Accessor

    TODO: THis should probablry bre refactored to a more generic download accessor

    """

    _dataset: Any = None

    def download(self, file_name: str | None = None) -> str:
        """Download the file from the presigned URL

        Args:
            file_name (str | None): The name of the file to save the downloaded file as
                If None, the key from the presigned URL will be used as the file name.
        Returns:
            str: The path to the downloaded file
        """
        from mikro_next.io.download import download_file

        url, key = get_attributes_or_error(self, "presigned_url", "key")
        return download_file(url, file_name=file_name or key)


class Vector(Protocol):
    """A Protocol for Vectorizable data

    Attributes:
        x (float): The x value
        y (float): The y value
        z (float): The z value
        t (float): The t value
        c (float): The c value
    """

    x: float
    y: float
    z: float
    t: float
    c: float

    def __init__(
        self,
        x: Optional[int] = None,
        y: Optional[int] = None,
        z: Optional[int] = None,
        t: Optional[int] = None,
        c: Optional[int] = None,
    ) -> None: ...


T = TypeVar("T", bound=Vector)


class HasPixelSizeTrait:
    """Mixin for PixelTranslatable data"""

    @property
    def pixel_size(self) -> Tuple[float, float, float]:
        """The pixel size of the representation

        Returns:
            Tuple[float, float, float]: The pixel size
        """
        kind, matrix = get_attributes_or_error(self, "kind", "matrix")

        if kind == "AFFINE":
            return tuple(np.array(matrix).reshape(4, 4).diagonal()[:3])

        raise NotImplementedError(f"Pixel size not implemented for this kind {kind}")

    @property
    def position(self) -> Tuple[float, float, float]:
        """The pixel size of the representation

        Returns:
            Tuple[float, float, float]: The pixel size
        """
        kind, matrix = get_attributes_or_error(self, "kind", "matrix")

        if kind == "AFFINE":
            return tuple(np.array(matrix).reshape(4, 4)[:3, 3])

        raise NotImplementedError(f"Pixel size not implemented for this kind {kind}")


class HasFromNumpyArrayTrait:
    """Mixin for Vectorizable data
    adds functionality to convert a numpy array to a list of vectors
    """

    @classmethod
    def list_from_numpyarray(
        cls: Type[T],
        x: TwoDArray,
        t: Optional[int] = None,
        c: Optional[int] = None,
        z: Optional[int] = None,
    ) -> List[T]:
        """Creates a list of InputVector from a numpya array

        Args:
            vector_list (List[List[float]]): A list of lists of floats

        Returns:
            List[Vectorizable]: A list of InputVector
        """
        assert x.ndim == 2, "Needs to be a List array of vectors"
        if x.shape[1] == 4:
            return [cls(x=i[1], y=i[0], z=i[2], t=i[3], c=c) for i in x.tolist()]
        if x.shape[1] == 3:
            return [cls(x=i[1], y=i[0], z=i[2], t=t, c=c) for i in x.tolist()]
        elif x.shape[1] == 2:
            return [cls(x=i[1], y=i[0], t=t, c=c, z=z) for i in x.tolist()]
        else:
            raise NotImplementedError(
                f"Incompatible shape {x.shape} of {x}. List dimension needs to either be of size 2 or 3"
            )

    @classmethod
    def from_array(
        cls: Type[T],
        x: OneDArray,
    ) -> T:
        """Creates a InputVector from a numpy array"""
        assert x.ndim == 1, "Needs to be a 1D array of floats"
        return cls(x=x[4], y=x[3], z=x[2], t=x[1], c=x[0])


class FileTrait:
    """A trait for file-like objects that can be downloaded
    because they have a big file store attached to them.
    """

    def download(self, file_name: str | None = None) -> "str":
        """Download the file from the store

        Args:
            file_name (str | None): The name of the file to save the downloaded file as
                If None, the key from the store will be used as the file name.
        Returns:
            str: The path to the downloaded file
        """
        store: "HasPresignedDownloadAccessor" = get_attributes_or_error(self, "store")
        return store.download(file_name=file_name)
